/*
 * =====================================================================================
 * 
 *        Filename:  applet.c
 * 
 *     Description:  main of the Gnome VM Applet
 * 
 *         Version:  $Id: applet.c,v 1.3 2006/01/27 17:05:33 kzak Exp $
 *         Created:  14.12.2005 09:07:36 CET
 * 
 *          Author:  Karel Zak <kzak@redhat.com>
 * 
 * =====================================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <config.h>
#include <string.h>

#include <gtk/gtk.h>
#include <libgnome/libgnome.h>
#include <panel-applet.h>
#include <panel-applet-gconf.h>
#include <gconf/gconf-client.h>

#include "applet.h"


static void vm_about_cb(BonoboUIComponent *, VmAppletCtl *, const char * );
static void vm_help_cb(BonoboUIComponent *, VmAppletCtl *, const char * );
static void vm_prop_cb(BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb );

static void vm_ctl_destroy(VmAppletCtl *ctl);
static void vm_destroy_cb(GtkWidget *widget, VmAppletCtl *ctl);

static GtkWidget *vm_mitem_os(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_state(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_activity(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_id(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_name(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_maxmem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_usemem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_pusemem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_ncpus(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);
static GtkWidget *vm_mitem_cputime(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item);

static void vm_popmenu_update(VmAppletCtl *ctl);
static gint vm_update(VmAppletCtl *ctl);
static void vm_icon_update(VmAppletCtl *ctl, int sz);
static void vm_domain_free(VmAppletDomain *dom);
static char *vm_domainstate_to_string(virDomainState state);

static int instances;

static const BonoboUIVerb vm_menu_verbs [] = 
{
    BONOBO_UI_UNSAFE_VERB ("VmAppletProp",  vm_prop_cb),
	BONOBO_UI_UNSAFE_VERB ("VmAppletHelp",  vm_help_cb),
	BONOBO_UI_UNSAFE_VERB ("VmAppletAbout", vm_about_cb),
    BONOBO_UI_VERB_END
};

/* Applet Icons */
static const VmAppletStockIcon vm_stock_icons[] =
{
    { VM_STOCK_BLOCKED,  VM_ICONDIR "/vm-blocked.png" },  
    { VM_STOCK_PAUSED,   VM_ICONDIR "/vm-paused.png" },
    { VM_STOCK_RUNNING,  VM_ICONDIR "/vm-running.png" },  
    { VM_STOCK_SHUTDOWN, VM_ICONDIR "/vm-shutdown.png" }, 
    { VM_STOCK_SHUTOFF,  VM_ICONDIR "/vm-shutoff.png" }
};

#define VM_ICON  (VM_ICONDIR "/vm-applet.png")

/* Applet popup menu fields */
static VmAppletMenuDesc vm_menu_desc[] =
{
    { VM_DATA_ACTIVITY, "Activity icon",    "mitem_activity", vm_mitem_activity, TRUE  },
    { VM_DATA_ID,       "Id",               "mitem_id",     vm_mitem_id,         TRUE  },
    { VM_DATA_NAME,     "Name",             "mitem_name",   vm_mitem_name,       TRUE  },
    { VM_DATA_MAXMEM,   "Max memory",       "mitem_maxmem", vm_mitem_maxmem,     FALSE },
    { VM_DATA_USEDMEM,  "Used memory",      "mitem_usemem", vm_mitem_usemem,     FALSE },
    { VM_DATA_USEDMEM,  "Used memory in %", "mitem_pusemem",vm_mitem_pusemem,    TRUE  },
    { VM_DATA_NCPUS,    "Number of CPUs",   "mitem_ncpus",  vm_mitem_ncpus,      FALSE },
    { VM_DATA_OSTYPE,   "OS type",          "mitem_os",     vm_mitem_os,         FALSE },
    { VM_DATA_STATE,    "State",            "mitem_state",  vm_mitem_state,      FALSE },
    { VM_DATA_CPUTIME,  "CPU Time",         "mitem_cputime",vm_mitem_cputime,    FALSE }
};

/* 
 * Prints and deallocates GError 
 */
static void
hmm_error(GError **error)
{
    _F();
	fprintf(stderr, "%s\n", (*error)->message);
	g_error_free (*error);
    *error = NULL;
}

/*
 * Initializes icons stock
 */
static void  
vm_init_icons(void)
{
    GtkIconFactory *icon_factory;
    gint i;

    _F();
    
	icon_factory = gtk_icon_factory_new();

	for (i = 0; i < G_N_ELEMENTS(vm_stock_icons); i++) 
    {
		VmAppletStockIcon icon = vm_stock_icons[i];
		GtkIconSource *icon_source = gtk_icon_source_new();
		GtkIconSet *icon_set = gtk_icon_set_new();

		gtk_icon_source_set_filename(icon_source, icon.filename);
		gtk_icon_set_add_source(icon_set, icon_source);
		gtk_icon_factory_add(icon_factory, icon.stock_id, icon_set);
		gtk_icon_source_free(icon_source);
		gtk_icon_set_unref(icon_set);
	}

	gtk_icon_factory_add_default(icon_factory);
	g_object_unref(G_OBJECT(icon_factory));
    
    gtk_window_set_default_icon_name("vm-applet");
}

/* 
 * Pops up an error dialog on the same screen as 'applet' saying 'msg'.
 */
static void
vm_error_dialog(GtkWidget *applet, char *msg)
{
    GtkWidget *dialog;
    
    _F();
    
    dialog = gtk_message_dialog_new( NULL, 0, GTK_MESSAGE_ERROR,
                                     GTK_BUTTONS_OK, msg);
    
    gtk_window_set_screen(GTK_WINDOW (dialog),
                           gtk_widget_get_screen (applet));
    
    gtk_window_set_title(GTK_WINDOW(dialog), "VM Applet");
    
    g_signal_connect_swapped( GTK_OBJECT (dialog), "response",
                              G_CALLBACK (gtk_widget_destroy),
                              GTK_OBJECT (dialog) );
    gtk_widget_show_all( dialog );
}

/* 
 * Common function invoked by the 'Help' context menu item and the 'Help'
 * button in the preferences dialog.
 */
void
vm_show_help( VmAppletCtl *ctl, const char *section )
{
    GError *error = NULL;
    
    _F();
    
    gnome_help_display_on_screen( "vm-applet", section,
                                gtk_widget_get_screen( ctl->applet ),
                                &error);
    if( error )
    {
        char *message;

        message = g_strdup_printf( _("There was an error displaying help: %s"),
                                 error->message );
        vm_error_dialog (ctl->applet, message);
        g_error_free( error );
        g_free( message );
    }
}

/* 
 * Called when the user selects the 'help' menu item.
 */
static void
vm_help_cb( BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb )
{
    _F();
    vm_show_help( ctl, NULL );
}

/* 
 * Called when the user selects the 'about' menu item.
 */
static void
vm_about_cb( BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb )
{
    const gchar *authors[] = {
        "Karel Zak <kzak@redhat.com>", 
        NULL
    };
    
    _F();
    
    gtk_show_about_dialog( NULL,
        "name",                _("VM Applet"), 
        "version",             VERSION,
        "copyright",           "\xC2\xA9 2005, 2006 Red Hat Inc. ",
        "comments",            _("This utility monitoring locally running Virtual domains."),
        "authors",             authors,
        "translator-credits",  _("translator-credits"),
        "logo-icon-name",      "vm-applet",
        NULL );
}

/* 
 * Called when the user selects the 'properties' menu item.
 */
static void
vm_prop_cb( BonoboUIComponent *uic, VmAppletCtl *ctl, const char *verb )
{
    _F();
    vm_prop_show(ctl);
}
   
/* 
 * Called when the user resizes the gnome panel
 */ 
static void
vm_size_cb( PanelApplet *applet, GtkAllocation *allocation, VmAppletCtl *ctl)
{
    vm_icon_update(ctl, MIN(allocation->width, allocation->height));    
}
    
/*
 * Creates applet control struct
 */
static VmAppletCtl *
vm_ctl_new(PanelApplet *applet)
{
    GError *error = NULL;
    VmAppletCtl *ctl;
    int i;

    _F();
    
    /* allocate ctl */
    ctl = g_new0(VmAppletCtl, 1);
    ctl->applet = GTK_WIDGET(applet);
    ctl->uid = getuid();
    ctl->instance = instances;
    panel_applet_add_preferences(applet, "/schemas/apps/vm-applet/prefs", NULL);
    
    /* delay */
    ctl->delay = panel_applet_gconf_get_int(applet, "delay", &error);
    if (error)
        hmm_error(&error);
    _D("read gconf: delay=%d", ctl->delay);
    ctl->delay = ctl->delay <= 0 ? 1000 : ctl->delay;
   
    /* initialize menu description */
    for (i=0; i < G_N_ELEMENTS(vm_menu_desc); i++)
    {
        VmAppletMenuDesc *desc = &vm_menu_desc[i];
        gboolean val;
      
        desc->ctl = ctl;
        
        val = panel_applet_gconf_get_bool(applet, desc->gconf, &error);
        _D("read gconf: %s=%d", desc->gconf, val);

        if (error)
            hmm_error(&error);
        else
            desc->enable = val; 
    }
    
    ctl->desc = vm_menu_desc;
    ctl->ndescs = G_N_ELEMENTS(vm_menu_desc);
    
    /* open connection */
    if (ctl->uid==0)
        ctl->conn = virConnectOpen(NULL);
    else
        ctl->conn = virConnectOpenReadOnly(NULL);
    
    return ctl;
}

/*
 * (Re)sets applet timer
 */
void
vm_timer_refresh(VmAppletCtl *ctl)
{
    _F();
    
	if (ctl->pixtimer) 
        g_source_remove(ctl->pixtimer);

    _D("setting delay to: %d (instance: %d)", ctl->delay, ctl->instance);
    ctl->pixtimer = g_timeout_add (ctl->delay, (GSourceFunc) vm_update, ctl);
}

/*
 * Removes (but not deallocates) domain frm list of domains
 */
static VmAppletDomain *
vm_domainslist_rem(VmAppletDomain *list, VmAppletDomain *dom)
{
    VmAppletDomain *d = list;
    
    _F();
         
    if (!list)
        return NULL;
    
    if (dom==list)
        return list->next;
    
    while(d)
    {
        if (d->next==dom)
        {
            d->next = dom->next;
            dom->next = NULL;
            break;
        }
        d = d->next;
    }
    return list;
}

/*
 * Adds domain to the list of domains
 */
static VmAppletDomain *
vm_domainslist_add(VmAppletDomain *list, VmAppletDomain *dom)
{
    VmAppletDomain *d = list;
    
    _F();

    dom->next = NULL;
    
    if (!list)
        return dom;
    
    while(d)
    {
        if (!d->next)
        {
            d->next = dom;
            break;
        }
        d = d->next;
    }
    return list;
}

/*
 * Looks up domain in the list of domains
 */
static VmAppletDomain *
vm_domainslist_lookup(VmAppletDomain *list, int id)
{
    _F();

    for (; list; list=list->next)
    {
        if (list->id==id)
            return list;
    }
    return NULL;
}

/*
 * Deallocates all domains in the list
 */
static void
vm_domainslist_free(VmAppletDomain *list)
{
    VmAppletDomain *d = list;

    _F();
    
    while(d)
    {
        VmAppletDomain *next = d->next;
        vm_domain_free(d);
        d = next;
    }
}
 
/*
 * Sets real state of domain
 * -- we assume the domain that uses CPU is not blocked. We cannot
 *    use only the domain->state field, because it's alwais "blocked"
 *    on sigle CPU boxes where is just running this code.
 */ 
static void
vm_domain_set_activity(VmAppletCtl *ctl, VmAppletDomain *dom)
{
    virDomainState old = dom->activity;
    
    _F();
    
    if (dom->state==VIR_DOMAIN_BLOCKED) 
    { 
        if ((ctl->delay * dom->cputime_ct) > VM_BLOCKED_TIME) 
            dom->activity = VIR_DOMAIN_BLOCKED; 
        else 
            dom->activity = VIR_DOMAIN_RUNNING; 
    } 
    else if (dom->state==VIR_DOMAIN_NOSTATE)
        dom->activity = old;                    /* better old state than nothing */
    else
        dom->activity = dom->state;             /* other than running or blocked state */


    _D("domain-activity: %s: '%s'", dom->name, vm_domainstate_to_string(dom->activity));

    if (old != dom->activity)
    {
        _D("domain-changed: %s: activity", dom->name);
        dom->changed |= VM_DATA_ACTIVITY;
    }
}

/*
 * Allocates new domain struct and load info about the domain
 */
static VmAppletDomain *
vm_domain_new(VmAppletCtl *ctl, virDomainPtr vir)
{
    VmAppletDomain *dom;
    virDomainInfo info;
    char *p;
    
    _F();
    
    dom = g_new0(VmAppletDomain, 1);
    dom->ctl = ctl;
    dom->changed = 0;
    
    dom->id = virDomainGetID(vir);
    dom->changed |= VM_DATA_ID;
    
    dom->name = g_strdup(virDomainGetName(vir));
    dom->changed |= VM_DATA_NAME;
    
    _D("domain-new: %s", dom->name);
    
    if ((p = virDomainGetOSType(vir)))
    {
        dom->ostype = g_strdup(p);
        dom->changed |= VM_DATA_OSTYPE;
        free(p);
    }
    if (virDomainGetInfo(vir, &info)==0)
    {
        dom->state = info.state;
        dom->changed |= VM_DATA_STATE;
        
        dom->maxmem = info.maxMem;
        dom->changed |= VM_DATA_MAXMEM;
        
        dom->usemem = info.memory;
        dom->changed |= VM_DATA_USEDMEM;
        
        dom->ncpus = info.nrVirtCpu;
        dom->changed |= VM_DATA_NCPUS;
        
        dom->cputime = info.cpuTime;
        dom->changed |= VM_DATA_CPUTIME;
    }
    
    vm_domain_set_activity(ctl, dom);
    return dom;
}

/*
 * Refreshs information about the domain
 */
static void
vm_domain_refresh(VmAppletCtl *ctl, VmAppletDomain *dom, virDomainPtr vir)
{
    virDomainInfo info;
    const char *p;
    
    _F();
    
    _D("domain-refresh: %s", dom->name);
    
    dom->changed = 0;
    
    if (dom->id!=virDomainGetID(vir))
    {
        _D("domain-changed: %s: id", dom->name);
        dom->id = virDomainGetID(vir);
        dom->changed |= VM_DATA_ID;
    }
    p = virDomainGetName(vir);
    if (p && strcmp(dom->name, p))
    {
        _D("domain-changed: %s: name", dom->name);
        g_free(dom->name);
        dom->name = g_strdup(p);
        dom->changed |= VM_DATA_NAME;
    }
    p = virDomainGetOSType(vir);
    if (p && strcmp(dom->ostype, p))
    {
        _D("domain-changed: %s: ostype", dom->name);
        g_free(dom->ostype);
        dom->ostype = g_strdup(p);
        dom->changed |= VM_DATA_OSTYPE;
    }
    if (p)
        free((void *)p);
    
    if (virDomainGetInfo(vir, &info)==0) 
    {
        if (dom->state != info.state)
        {
            _D("domain-changed: %s: state", dom->name);
            dom->state = info.state;
            dom->changed |= VM_DATA_STATE;
        }
        if (dom->maxmem != info.maxMem)
        {
            _D("domain-changed: %s: maxmem", dom->name);
            dom->maxmem = info.maxMem;
            dom->changed |= VM_DATA_MAXMEM;
        }
        if (dom->usemem != info.memory)
        {
            _D("domain-changed: %s: usemem", dom->name);
            dom->usemem = info.memory;
            dom->changed |= VM_DATA_USEDMEM;
        }
        if (dom->ncpus != info.nrVirtCpu)
        {
            _D("domain-changed: %s: ncpus", dom->name);
            dom->ncpus = info.nrVirtCpu;
            dom->changed |= VM_DATA_NCPUS;
        }
        if (dom->cputime != info.cpuTime)
        {
            _D("domain-changed: %s: cputime", dom->name);
            dom->cputime = info.cpuTime;
            dom->changed |= VM_DATA_CPUTIME;
            dom->cputime_ct = 0;
        }
        else
            dom->cputime_ct++; 
    }
    vm_domain_set_activity(ctl, dom);
}

/*
 * Updates all domains 
 */
static void
vm_domains_update(VmAppletCtl *ctl)
{
    VmAppletDomain *old, *new, *dom;
    int *ids, i;

    _F();
    
    if (!ctl->conn)
        return;
    
    old = ctl->dom;
    new = NULL;
    
    ctl->ndoms = virConnectNumOfDomains(ctl->conn);
    if (ctl->ndoms <= 0)
    {
       vm_domainslist_free(ctl->dom);
       ctl->dom = NULL;
       return;
    }
    ids = g_malloc(sizeof(int) * ctl->ndoms);
    virConnectListDomains(ctl->conn, &ids[0], ctl->ndoms);

    for(i=0; i < ctl->ndoms; i++) 
    {
        virDomainPtr vir = virDomainLookupByID(ctl->conn, ids[i]);
        int id = virDomainGetID(vir);
        
        if ((dom = vm_domainslist_lookup(old, id)))
        {
            old = vm_domainslist_rem(old, dom);
            vm_domain_refresh(ctl, dom, vir);
        }
        else
            dom =  vm_domain_new(ctl, vir);
        
        new = vm_domainslist_add(new, dom);
        virDomainFree(vir);
    }

    if (old)
        vm_domainslist_free(old);
    ctl->dom = new;
}

/*
 * Deallocate the domain
 */
static void
vm_domain_free(VmAppletDomain *dom)
{
    _F();
    
    if (!dom)
        return;
    
    if (dom->ctl->popmenu && dom->menuitem) 
        gtk_widget_destroy(dom->menuitem);
    if (dom->name)
        g_free(dom->name);
    if (dom->ostype)
        g_free(dom->ostype);

    g_free(dom);
}

/*
 * Deallocates the applet control struct
 */
static void
vm_ctl_destroy(VmAppletCtl *ctl)
{
    _F();
    
    if (ctl->dom)
        vm_domainslist_free(ctl->dom);
    if (ctl->conn)
        virConnectClose(ctl->conn);

    g_free(ctl);
}

/*
 * Updates the tool tip on the gome-panel
 */
static void
vm_tooltips_update(VmAppletCtl *ctl)
{
	char *tooltip;
	
    _F();
    
    if (ctl->ndoms > 0)
    {
        if (ctl->ndoms > 1)
        	tooltip = g_strdup_printf (_("%d virtual domains"), ctl->ndoms);
        else
            tooltip = g_strdup_printf (_("%d virtual domain"), ctl->ndoms);
        gtk_tooltips_set_tip (ctl->applet_tip, ctl->applet, 
                tooltip, NULL);
        g_free (tooltip);
    }
    else if (!ctl->conn)
        gtk_tooltips_set_tip (ctl->applet_tip, ctl->applet, 
                _("Failed to connect to the hypervisor."), NULL);
    else
        gtk_tooltips_set_tip (ctl->applet_tip, ctl->applet, 
                _("Failed to load information about virtual domains."), NULL);
}

/* 
 * Creates or updates the applet 
 */
static gint
vm_update(VmAppletCtl *ctl)
{
    _D("\n\n---- loop (pid: %d, instance: %d, delay: %d) ----", getpid(), ctl->instance, ctl->delay);
    _F();
    
    if (!ctl->conn)
    {
        if (ctl->uid==0)
            ctl->conn = virConnectOpen(NULL);
        else
            ctl->conn = virConnectOpenReadOnly(NULL);
    }
    
    vm_domains_update(ctl);
    
    if (ctl->popmenu && ctl->popmenu_up==1)
        vm_popmenu_update(ctl);

    vm_tooltips_update(ctl);
    
    _D("\n\n");
    
    return TRUE;
}

/*
 * Deallocates the applet (the end! ;-)
 */
static void
vm_destroy_cb(GtkWidget *widget, VmAppletCtl *ctl)
{   
    _F(); 
    
    g_source_remove(ctl->pixtimer);
    
    if (ctl->popmenu)
    {
        gtk_widget_destroy(ctl->popmenu);
        ctl->popmenu = NULL;
    }
    vm_ctl_destroy(ctl);

    --instances;
}


/* 
 * Counts pop menu position
 * -- the following function is adapted from gtkmenuitem.c 
 */
static void
vm_position_popmenu (GtkMenu *menu, gint *x, gint *y,
	       gboolean *push_in, gpointer ctl)
{
    GtkWidget *widget = ((VmAppletCtl *) ctl)->applet;
    GdkScreen *screen;
    gint twidth, theight, tx, ty;
    GtkTextDirection direction;
    GdkRectangle monitor;
    gint monitor_num;

    _F();
    
    g_return_if_fail (menu != NULL);
    g_return_if_fail (x != NULL);
    g_return_if_fail (y != NULL);

    if (push_in) *push_in = FALSE;

    direction = gtk_widget_get_direction (widget);

    twidth = GTK_WIDGET (menu)->requisition.width;
    theight = GTK_WIDGET (menu)->requisition.height;

    screen = gtk_widget_get_screen (GTK_WIDGET (menu));
    monitor_num = gdk_screen_get_monitor_at_window (screen, widget->window);
    if (monitor_num < 0) monitor_num = 0;
    gdk_screen_get_monitor_geometry (screen, monitor_num, &monitor);

    if (!gdk_window_get_origin (widget->window, &tx, &ty)) 
    {
	    g_warning ("Menu not on screen");
	    return;
    }

    tx += widget->allocation.x;
    ty += widget->allocation.y;

    if (direction == GTK_TEXT_DIR_RTL)
	tx += widget->allocation.width - twidth;

    if ((ty + widget->allocation.height + theight) <= monitor.y + monitor.height)
	ty += widget->allocation.height;
    else if ((ty - theight) >= monitor.y)
	ty -= theight;
    else if (monitor.y + monitor.height - (ty + widget->allocation.height) > ty)
	ty += widget->allocation.height;
    else
	ty -= theight;

    *x = CLAMP (tx, monitor.x, MAX (monitor.x, monitor.x + monitor.width - twidth));
    *y = ty;
    gtk_menu_set_monitor (menu, monitor_num);
}

/*
 * Called when the user activates the applet menu item (TODO)
 */
static void
vm_menuitem_cb (GtkMenuItem *menuitem, gpointer user_ctl)
{
    /*VmAppletDomain *domain = (VmAppletDomain *) user_ctl;*/

    _F();
    ;
}

/*
 * Blocked or really dead?
 */
#define VM_IS_BLOCKED(_ctl, _dom) \
            ((_ctl)->delay * (_dom)->cputime_ct > VM_BLOCKED_TIME)

/*
 * Converts the domain state to string
 */
static char *
vm_domainstate_to_string(virDomainState state) 
{
    _F();
    switch (state) {
        case VIR_DOMAIN_BLOCKED:
            return _("blocked");
        case VIR_DOMAIN_RUNNING:
            return _("running");
        case VIR_DOMAIN_PAUSED:
            return _("paused");
        case VIR_DOMAIN_SHUTDOWN:
            return _("in shutdown");
        case VIR_DOMAIN_SHUTOFF:
            return _("shut off");
        default:
            return NULL;    /* = dom0 state */
    }
    return NULL;
}

/*
 * Converts the domain state to icon stock Id
 */
static char *
vm_domainstate_to_stockid(virDomainState state) 
{
    _F();
    switch (state) {
        case VIR_DOMAIN_BLOCKED:
            return VM_STOCK_BLOCKED;
        case VIR_DOMAIN_RUNNING:
            return VM_STOCK_RUNNING;
        case VIR_DOMAIN_PAUSED:
            return VM_STOCK_PAUSED;
        case VIR_DOMAIN_SHUTDOWN:
            return VM_STOCK_SHUTDOWN;
        case VIR_DOMAIN_SHUTOFF:
            return VM_STOCK_SHUTOFF;
        default:
            return NULL;    /* = dom0 state */
    }
    return NULL;
}

static GtkWidget *
vm_mitem_os(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    _F();

    if (item)
        gtk_label_set_text(GTK_LABEL(item), domain->ostype);
    else
    {
        item = gtk_label_new(domain->ostype);
        gtk_misc_set_alignment (GTK_MISC(item), 0.0, 0.5);
    }
    return item;
}


static GtkWidget *
vm_mitem_state(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    _F();

    if (item)
        gtk_label_set_text(GTK_LABEL(item), vm_domainstate_to_string(domain->state));
    else
    {
        item = gtk_label_new(vm_domainstate_to_string(domain->state));
        gtk_misc_set_alignment (GTK_MISC(item), 0.0, 0.5);
    }
    return item;
}

static GtkWidget *
vm_mitem_activity(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *id;
    _F();

    if (!(id = vm_domainstate_to_stockid(domain->activity)))
    {
        if (domain->id==0)      /* dom0 */
            id = vm_domainstate_to_stockid(VIR_DOMAIN_RUNNING);
        else
            id = "";        /* TODO: create unknown state icon */
    }
   
    if (item)
        gtk_image_set_from_stock(GTK_IMAGE(item), id, GTK_ICON_SIZE_MENU); 
    else
    {
        item =  gtk_image_new_from_stock (id, GTK_ICON_SIZE_MENU);
        gtk_misc_set_alignment (GTK_MISC(item), 0.5, 0.5);
    }
    return item;
}

static GtkWidget *
vm_mitem_id(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str;
    _F();

    str = g_strdup_printf("%d", domain->id);
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_name(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    _F();

    if (item)
        gtk_label_set_text(GTK_LABEL(item), domain->name);
    else
    {
        item = gtk_label_new(domain->name);
        gtk_misc_set_alignment (GTK_MISC(item), 0.0, 0.5);
    }
    return item;
}

static GtkWidget *
vm_mitem_maxmem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str;
    _F();

    str = g_strdup_printf("%lu MB", (unsigned long)(domain->maxmem / (double)1024));
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_usemem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str;
    _F();

    str = g_strdup_printf("%lu MB", (unsigned long)(domain->usemem / (double)1024));
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_pusemem(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str;
    _F();

    str = g_strdup_printf("%d%%", (int)(domain->usemem / (domain->maxmem / (double)100)));
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str);
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_ncpus(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str;
    _F();

    str = g_strdup_printf("%d %s", domain->ncpus, domain->ncpus > 0 ? "CPUs" : "CPU");
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str); 
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

static GtkWidget *
vm_mitem_cputime(VmAppletDomain *domain, VmAppletMenuDesc *desc, GtkWidget *item)
{
    char *str = NULL;
    float cputime = domain->cputime;

    _F();

    if (cputime != 0) 
    {
        cputime /= 1000000000;
        str = g_strdup_printf("%.1fs", cputime);
    }
    if (item)
        gtk_label_set_text(GTK_LABEL(item), str); 
    else
    {
        item = gtk_label_new(str);
        gtk_misc_set_alignment (GTK_MISC(item), 1.0, 0.5);
    }
    g_free(str);
    return item;
}

/*
 * Creates menu item
 */
static GtkWidget *
vm_popmenuitem_new(VmAppletCtl *ctl, VmAppletDomain *dom)
{
    int i;
    GtkWidget *box;
    GtkWidget *menuitem;

    _F();
    
    menuitem = gtk_menu_item_new ();
    gtk_widget_set_sensitive (menuitem, TRUE);
    gtk_container_add (GTK_CONTAINER (ctl->popmenu), menuitem);
    g_signal_connect(G_OBJECT(menuitem), "activate", G_CALLBACK(vm_menuitem_cb), dom);

    box = gtk_hbox_new(FALSE, 12);
    gtk_container_add (GTK_CONTAINER(menuitem), box);
            
    for(i=0; i < ctl->ndescs; i++)
    {
        VmAppletMenuDesc *desc = &ctl->desc[i];  
        GtkWidget *subitem;
        
        if (!desc->enable)
            continue;
        
        subitem = desc->builder(dom, desc, NULL);
        if (!subitem)
        {
            _D("popmenu-item-new: %s: no item: %s", dom->name, desc->name);
            subitem = gtk_label_new("");    /* dummy column */
        }
        gtk_box_pack_start(GTK_BOX(box), subitem, FALSE, FALSE, 0);
        gtk_size_group_add_widget(desc->sizegroup, subitem);
    }
    gtk_widget_show_all (menuitem);
    
    return menuitem;
}

/*
 * Updates fields in the menu item
 */
static GtkWidget *
vm_popmenuitem_update(VmAppletCtl *ctl, VmAppletDomain *dom, GtkWidget *menuitem)
{
    GtkWidget *box;
    GList *list, *li;
    int i=0;

    
    _F();
    
    if (!dom->changed)
        return menuitem;
    
    /* get Box */
    if (!(list = gtk_container_get_children(GTK_CONTAINER(menuitem))))
        return menuitem;
    li = g_list_first(list);
    box = li->data;
    g_list_free(list);
    
    /* get Items */
    list = gtk_container_get_children(GTK_CONTAINER(box));
    _D("popmenu-item-update: %s: found: %d subitems", dom->name, list ? g_list_length(list) : 0);
    
    li = g_list_first(list);
    for(i=0; i < ctl->ndescs; i++)
    {
        VmAppletMenuDesc *desc = &ctl->desc[i];  
        GtkWidget *subitem;
                
        if (!desc->enable)
        {
            _D("popmenu-item-disabled: %s: %s", dom->name, desc->name);
            continue;
        }
        if (li==NULL)
        {
            _D("*** out of list! ***");
            break;
        }
        subitem = li->data;
        li = g_list_next(li);
        
        if (!(dom->changed & desc->did))
        {
            _D("popmenu-item-nochange: %s: %s", dom->name, desc->name);
            continue;
        }
        _D("popmenu-item-change: %s: %s", dom->name, desc->name);
        subitem = desc->builder(dom, desc, subitem);
    }
    return menuitem;
}

/*
 * Updates all items in the menu
 */
static void
vm_popmenu_update(VmAppletCtl *ctl)
{
    VmAppletDomain *dom;
    
    _F();
    
    for(dom=ctl->dom; dom; dom=dom->next)
    {
        if (dom->menuitem==NULL)
        {
            dom->menuitem = vm_popmenuitem_new(ctl, dom);
        }
        else if (dom->changed==0)
        {
            _D("popmenu-nochange: %s", dom->name);
            continue;
        }
        else 
        {
            _D("popmenu-change: %s", dom->name);
            vm_popmenuitem_update(ctl, dom, dom->menuitem);
        }
        dom->changed = 0;
    }
}

void
vm_popmenu_destroy_items(VmAppletCtl *ctl)
{
    VmAppletDomain *dom;
    
    _F();
    
    for(dom=ctl->dom; dom; dom=dom->next)
    {
        if (dom->menuitem)
            gtk_widget_destroy(dom->menuitem);
        dom->menuitem = NULL;
    }
}

/*
 * Called when the menu pop down
 */
static void
vm_popmenu_deactivate_cb(GtkMenuShell *menushell, VmAppletCtl *ctl)
{
    _F();
    ctl->popmenu_up = 0;
}

/*
 * Creates the applet pop menu
 */ 
static void
vm_popmenu_new(VmAppletCtl *ctl)
{
    int i;
   
    _F();
       
    if (ctl->popmenu)
        return;
    
    ctl->popmenu = gtk_menu_new ();
    ctl->popmenu_up = 0;

    for(i=0; i < ctl->ndescs; i++)
        ctl->desc[i].sizegroup = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

	g_signal_connect(G_OBJECT(ctl->popmenu), "deactivate",
			G_CALLBACK(vm_popmenu_deactivate_cb), ctl);

    vm_popmenu_update(ctl);
}

/*
 * Called when user press the button on the gnome applet
 */
gboolean
vm_button_cb(   GtkWidget *widget, 
                        GdkEventButton *event, 
                        VmAppletCtl *ctl)
{
    _F();

    if (event->button != 1)
        return FALSE;
    
    if (!ctl->popmenu)
        vm_popmenu_new(ctl);
    else 
        vm_popmenu_update(ctl);
    
    ctl->popmenu_up = 1;
    
	gtk_menu_popup (GTK_MENU (ctl->popmenu), NULL, NULL,
			    vm_position_popmenu, ctl,
			    event->button, event->time);
    
    return TRUE;
}

/*
 * Updates the incon on the gnome panel
 * -- called when user resizes the panel
 */
static void
vm_icon_update(VmAppletCtl *ctl, int sz)
{
    GdkPixbuf *pixbuf;
    
    _F();
    
    if (sz==ctl->image_size)
        return;
    
    ctl->image_size = sz;
    
    if (!ctl->image)
    {
        GError *error = NULL;
        
        ctl->pixbuf = gdk_pixbuf_new_from_file(VM_ICON, &error);
        if (error)
            hmm_error(&error);
       
        ctl->image = gtk_image_new();
        gtk_container_add(GTK_CONTAINER(ctl->applet), ctl->image);
    }
    
    pixbuf = gdk_pixbuf_scale_simple(ctl->pixbuf, sz, sz, GDK_INTERP_BILINEAR);
    gtk_image_set_from_pixbuf(GTK_IMAGE(ctl->image), pixbuf);

    g_object_unref(pixbuf);
}

/* 
 * Called by the factory to fill in the fields for the applet.
 */
static gboolean
vm_setup (PanelApplet *applet)
{
    VmAppletCtl *ctl;
    AtkObject *atk;
    
    _F();

    if (instances==1)
        vm_init_icons();

    panel_applet_set_flags(applet, PANEL_APPLET_EXPAND_MINOR); 

    ctl = vm_ctl_new(applet);
    
    vm_icon_update(ctl, panel_applet_get_size(PANEL_APPLET(ctl->applet)));
    
    panel_applet_setup_menu_from_file ( PANEL_APPLET (applet), 
      	        		                DATADIR,
                                        "GNOME_VmApplet.xml",
                                        NULL,
                                        vm_menu_verbs,
                                        ctl);

    if (panel_applet_get_locked_down (PANEL_APPLET (applet))) 
    {
        BonoboUIComponent *popup_component;

	    popup_component = panel_applet_get_popup_component (PANEL_APPLET (applet));
        bonobo_ui_component_set_prop (popup_component,
					"/commands/preferences",
					"hidden", "1",
					NULL);
    }

	g_signal_connect(G_OBJECT(ctl->applet), "button-press-event",
			G_CALLBACK(vm_button_cb), ctl);
    g_signal_connect(G_OBJECT(ctl->applet), "destroy",
			G_CALLBACK(vm_destroy_cb), ctl);
    g_signal_connect(G_OBJECT(ctl->applet), "size_allocate",
   		    G_CALLBACK(vm_size_cb), ctl);
    
    atk = gtk_widget_get_accessible (ctl->applet);
    if (GTK_IS_ACCESSIBLE (atk)) 
    {
        atk_object_set_name (atk, _("VmApplet"));
        atk_object_set_description(atk, _("Monitor locally running virtual domains"));
    }    
    
    gtk_widget_show_all(ctl->applet);
    
    ctl->applet_tip =  gtk_tooltips_new();
   
    vm_update(ctl);
    vm_timer_refresh(ctl); 
       
    return TRUE;
}

/*
 * "main()" of the applet
 */
static gboolean
vm_factory(PanelApplet *applet, const gchar *iid, gpointer ptr)
{
	gboolean retval = FALSE;
    
    instances++;
    _D("---- INITIALIZE instance: %d ----", instances);
    
    _F();
    
    if (!strcmp(iid, "OAFIID:GNOME_VmApplet")) 
        retval = vm_setup(applet);

    return retval;
}

/*
 * Bonobo initialization
 */
PANEL_APPLET_BONOBO_FACTORY("OAFIID:GNOME_VmApplet_Factory",
                             PANEL_TYPE_APPLET,
                             "vm-applet",
                             "0",
                             (PanelAppletFactoryCallback) vm_factory,
                             NULL)


/*
 * vim: set tabstop=4:
 * vim: set shiftwidth=4:
 * vim: set expandtab:
 */
