/*
 * =====================================================================================
 * 
 *        Filename:  applet.h
 * 
 *         Version:  $Id: applet.h,v 1.3 2006/02/10 11:21:25 veillard Exp $
 *         Created:  15.12.2005 20:22:13 CET
 * 
 *          Author:  Karel Zak <kzak@redhat.com>
 * 
 * =====================================================================================
 */

#ifndef __APPLET_H__
#define __APPLET_H__

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gtk/gtksizegroup.h>
#include <gtk/gtktooltips.h>
#include <gtk/gtkwidget.h>
#include <panel-applet.h>

#include <libvirt/libvirt.h>

/*#define DEBUG   1*/

#ifdef DEBUG
#define _D(_format, _a...)      fprintf(stderr, _format "\n", ##_a)
#define _F()                    fprintf(stderr, "--> %s()\n", __FUNCTION__)
#else
#define _D(_format, _a...)
#define _F()
#endif

/* icon stock names */
#define VM_STOCK_BLOCKED	"vmapplet-blocked"
#define VM_STOCK_PAUSED		"vmapplet-paused"
#define VM_STOCK_RUNNING	"vmapplet-running"
#define VM_STOCK_SHUTDOWN	"vmapplet-shutdown"
#define VM_STOCK_SHUTOFF	"vmapplet-shutoff"

/* dummy */
typedef struct _VmAppletCtl VmAppletCtl;

typedef struct
{
	gchar	*stock_id;
	gchar	*filename;
} VmAppletStockIcon;

/* the domain without CPU time change for VM_BLOCKED_TIME 
 * (milliseconds) is marked as "blocked"
 */
#define VM_BLOCKED_TIME		10

/* domain description */
typedef struct VmAppletDomain
{
	int			id;		/* domain ID */
	char			*name;		/* domain Name */
	char			*ostype;	/* domain OS Type */
	virDomainState		state;		/* domain State */
	virDomainState		activity;	/* real state by State + CpuTime */
	unsigned long		maxmem;		/* domain Max allocated Memory */
	unsigned long		usemem;		/* domain Used Memory */
	unsigned long long	cputime;	/* domain CPU Time */
	int			cputime_ct;	/* number of loops without CPU Time change */
	int 			ncpus;		/* domain number of CPUs */
	GtkWidget		*menuitem;	/* popup menu item */
	VmAppletCtl		*ctl;		/* global applet control struct */
	int			changed;	/* bit area with info about changed columns */
	struct VmAppletDomain	*next;		/* next domain */
} VmAppletDomain;

#define VM_DATA_ID       (1 << 1)
#define VM_DATA_NAME     (1 << 2)
#define VM_DATA_OSTYPE   (1 << 3)
#define VM_DATA_ACTIVITY (1 << 4)	/* real state */
#define VM_DATA_MAXMEM   (1 << 5)
#define VM_DATA_USEDMEM	 (1 << 6)
#define VM_DATA_NCPUS    (1 << 7)
#define VM_DATA_CPUTIME  (1 << 8)
#define VM_DATA_STATE    (1 << 9)

/* menu item description */
typedef struct VmAppletMenuDesc
{
	int		did;			/* data source ID */
	char		*name;			/* column name */
	char		*gconf;			/* pointer to gnome config */
						/* data builder */
	GtkWidget	*(*builder)(VmAppletDomain *, struct VmAppletMenuDesc *, GtkWidget *item);

	gboolean	enable;			/* visible? */
	GtkSizeGroup	*sizegroup;		/* GTK size group */
	VmAppletCtl	*ctl;			/* global applet control struct */
} VmAppletMenuDesc;

/* Internal applet ctl */
typedef struct _VmAppletCtl
{
	GtkWidget		*applet;	/* applet widget */
	GtkTooltips		*applet_tip;	/* applet tip */
	GtkWidget		*image;		/* applet icon */
	GdkPixbuf 		*pixbuf;	/* icon pixbuf */
	int			image_size;
	GtkWidget		*popmenu;	/* list of domais */
	GtkWidget		*propwin;	/* preferences window */
	int			popmenu_up;	/* is the pop menu UP/DOWN */
	VmAppletMenuDesc	*desc;		/* menu columns descriptions */
	int			ndescs;		/* num of items in "descs" */
	VmAppletDomain		*dom;		/* Vm domains */
	int			ndoms;		/* number of Vm domains */
	virConnectPtr		conn;		/* connection to hypervisor */
	guint  			pixtimer;	/* glib timer */
	int 			delay;		/* update delay */
	uid_t			uid;
	int			instance;
} _VmAppletCtl; 

extern void vm_show_help( VmAppletCtl *ctl, const char *section );
extern void vm_prop_show(VmAppletCtl *ctl);
extern void vm_timer_refresh(VmAppletCtl *ctl);
extern void vm_popmenu_destroy_items(VmAppletCtl *ctl);

#endif /* __APPLET_H__ */

